from numpy.random import choice, seed


def get_first_random_solution(m, data):
    seed(42)
    random_indexes = choice(len(data.index), size=m)
    return data.iloc[random_indexes]


def replace_worst_element(previous, data):
    solution = previous.copy()
    worst_index = solution["distance"].astype(float).idxmin()
    solution.loc[worst_index] = data.sample(random_state=42).squeeze()
    return solution, worst_index


def get_random_solution(previous, data):
    candidates = []
    candidates.append(previous)
    solution, worst_index = replace_worst_element(previous, data)
    previous_worst_distance = previous["distance"].loc[worst_index]
    while candidates[-1].distance.loc[worst_index] <= previous_worst_distance:
        last_solution = candidates[-1]
        solution, _ = replace_worst_element(previous=solution, data=data)
        if solution.equals(last_solution):
            return last_solution
        candidates.append(solution)
    return solution


def explore_neighbourhood(element, data, max_iterations=100000):
    neighbourhood = []
    neighbourhood.append(element)
    for _ in range(max_iterations):
        previous_solution = neighbourhood[-1]
        neighbour = get_random_solution(previous=previous_solution, data=data)
        if neighbour.equals(previous_solution):
            break
        neighbourhood.append(neighbour)
    return neighbour


def local_search(m, data):
    first_solution = get_first_random_solution(m=m, data=data)
    best_solution = explore_neighbourhood(element=first_solution, data=data)
    return best_solution
